// heap : { Handle : Namespace }        -- Map[Handle,Map[String,Meaning]]
// where
// Namespace = { Identifier : Meaning } -- Map[String,Meaning]
// Meaning = Handle | Int               -- 

// Program P ::= T
// CommandList CL ::= C | C ; CL
// Command C ::= var I = E | L = E | while E : CL end
// Expression E ::= N | ( E1 + E2 ) | L | new T
// Template T ::= { CL }
// LefthandSide L ::= I | L . I | this 
// Numeral N ::= string of digits
// Variable I ::= strings of letters, not including keywords: while, print, end

trait OpTree {
  sealed abstract class Ctree
  case class Decl(x: String, e:Etree) extends Ctree
  case class Assign(l: List[String], e: Etree) extends Ctree
  case class While(e: Etree, c: List[Ctree]) extends Ctree
  
  sealed abstract class Etree
  case class Num(s: String) extends Etree
  case class Add(e1: Etree, e2: Etree) extends Etree
  case class Sub(e1: Etree, e2: Etree) extends Etree
  case class At(l: List[String]) extends Etree
  case class New(t: Ttree) extends Etree
  
  sealed abstract class Ttree
  case class Template(cs: List[Ctree]) extends Ttree
}

import scala.util.parsing.combinator.JavaTokenParsers

object OOPLcore extends JavaTokenParsers with OpTree {

  // Parser
  
  def parse(source: String): Ttree = 
    parseAll(prog, source) match {
      case Success(optree,_) => optree
      case _ => throw new Exception("Parse error!")
    }
  
  def prog: Parser[Ttree] = template
  
  def commlist: Parser[List[Ctree]] = rep1sep(comm, ";")
  
  def comm: Parser[Ctree] = "var"~>ident~("="~>expr) ^^ 
                              { case x~e => Decl(x,e) } | 
                            left~("="~>expr) ^^ 
                              { case l~e => Assign(l,e) } |
                            ("while"~>expr<~":")~(commlist<~"end") ^^ 
                              { case e~cs => While(e,cs) }

  def expr: Parser[Etree] = wholeNumber ^^ (Num(_)) | 
                           "("~>expr~op~expr<~")" ^^
                           { case e1~"+"~e2 => Add(e1,e2)
                             case e1~"-"~e2 => Sub(e1,e2) } |
                           "new"~>template ^^ (New(_)) |  
                           left ^^ (At(_)) 

  def template: Parser[Ttree] = "{"~>commlist<~"}" ^^ 
                                { case cs => (Template(cs)) }

  def left: Parser[List[String]] = rep1sep(ident, ".")

  def op: Parser[String] = "+" | "-"

  // Namespace Algebra
  // heap : { Handle : Namespace }
  // where Namespace = { Identifier : Meaning }
  //       Meaning = Handle | Integer
  //       LValue = Handle * Identifier

  sealed abstract class Meaning
  case class Handle(loc: Int) extends Meaning
  case class Value(n: Int) extends Meaning

  var heap: Map[Handle, Map[String, Meaning]] = Map()

  def alloc(d: Map[String, Meaning]): Handle = {
    val n: Int = heap.size
    heap += (Handle(n) -> d) // allocate a new cell for the object d
    Handle(n) // return the handle of newly allocated object
  }

  def bind(handle: Handle, fieldname: String, meaning: Meaning): Handle = {
    if (heap contains handle) { 
      heap += (handle -> (heap(handle) + (fieldname -> meaning)))
      handle
    }
    else // if the handle is not in the heap
      throw new Exception("bind error: " + handle + " does not exist in the heap")
  }

  def find(handle: Handle, fieldname: String): Meaning = {
    if (heap contains handle) 
      heap(handle)(fieldname)
    else 
      throw new Exception("find error: " + handle + " does not exist in the heap")
  }

  def member(fieldname: String, handle: Handle): Boolean = {
    if (heap contains handle)
      heap(handle) contains fieldname
    else 
      throw new Exception("find error: " + handle + " does not exist in the heap")  
  }  
  
  // Activation Stack Algebra

  def push(h: Handle): Unit = {
    val newtop = alloc(Map("ns"->h, "parent"->actstack))
    actstack = newtop
  } 

  def pop(): Unit = {
    actstack = find(actstack, "parent") 
               match {
                 case Handle(loc) => Handle(loc)
                 case _ => throw new Exception("Must be a handle.")
               }
  } 

  def top(): Handle = {
    if (actstack == Handle(-1)) Handle(-1)
    else find(actstack, "ns") 
         match {
           case Handle(loc) => Handle(loc)
           case _ => throw new Exception("Must be an address.")
         }
  }

  // Interpreter
  
  var actstack: Handle = Handle(-1)
  
  def interpretPTREE(p: Ttree): Unit = interpretTTREE(p)
  
  def interpretCLIST(cs: List[Ctree]): Unit = 
    for (c <- cs) yield interpretCTREE(c)
  
  def interpretCTREE(c: Ctree): Unit = c match {
    case Decl(x,e) => bind(top(),x,interpretETREE(e))
    case Assign(l,e) => {
      val (handle,x) = interpretLLIST(l)
      bind(handle,x,interpretETREE(e))
    }
    case While(e,cs) => {
      val cond = interpretETREE(e)
      if (cond != 0) {
        interpretCLIST(cs)
        interpretCTREE(c)
      }
    }
  }
  
  def interpretETREE(e: Etree): Meaning = e match {
    case Num(n) => Value(n.toInt)
    case Add(e1,e2) => {
      val n1 = interpretETREE(e1) match {
                 case Handle(_) => throw new Exception
                 case Value(n) => n
               }
      val n2 = interpretETREE(e2) match {
                 case Handle(_) => throw new Exception
                 case Value(n) => n
               }
      Value(n1 + n2)
    }
    case Sub(e1,e2) => 
      {
      val n1 = interpretETREE(e1) match {
                 case Handle(_) => throw new Exception
                 case Value(n) => n
               }
      val n2 = interpretETREE(e2) match {
                 case Handle(_) => throw new Exception
                 case Value(n) => n
               }
      Value(n1 - n2)
    }
    case At(l) => {
      val (h,x) = interpretLLIST(l)
      find(h, x)
    } 
    case New(t) => interpretTTREE(t)
  }
  
  def interpretTTREE(t: Ttree): Handle = t match {
    case Template(cs) => {
      val newhandle = alloc(Map("parent"->top()))
      push(newhandle)
      interpretCLIST(cs)
      pop()
      newhandle
    }
  }
  
  def searchStatic(thishandle: Handle, x: String): (Handle, String) = {
    if (thishandle != Handle(-1)) {
      if (member(x, thishandle)) (thishandle,x)
      else find(thishandle, "parent") match {
        case Handle(loc) => searchStatic(Handle(loc),x)
        case _ => throw new Exception()
      }
    }
    else throw new Exception()
  }
  
  def interpretLLIST(lefts: List[String]): (Handle,String) = {
    def outsideIn(h: Handle, lefts: List[String]): (Handle, String) = {
      lefts match {
        case left::Nil => searchStatic(h, left)
        case left::lefts =>
          val (h1, newleft) = searchStatic(h, left)
          find(h1, newleft) match {
            case Handle(loc) => outsideIn(Handle(loc), lefts)
            case _ => throw new Exception()
          }
        case Nil => throw new Exception("This will never be reached.")
      }
    }
    lefts match {
      case Nil => throw new Exception()
      case "this"::Nil => (actstack, "ns")
      case left::Nil => searchStatic(top(),left)
      case "this"::lefts => outsideIn(top(), lefts)
      case lefts => outsideIn(top(), lefts)
    }
  }
  
  // Controller
  def main(args: Array[String]): Unit = {
    try {
      val source = args(0)
      println("input : " + source)
      val optree = parse(source)
      println("optree : " + optree)
      interpretPTREE(optree)
      println("actstack : " + actstack)
      println("heap : " + heap)
    }
    catch { case e: Exception => println(e) }
  }
}
   




